/*
 * This file is part of roccat-tools.
 *
 * roccat-tools is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 *
 * roccat-tools is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with roccat-tools. If not, see <http://www.gnu.org/licenses/>.
 */

#include "roccat_secure.h"
#include <stdlib.h>
#include <string.h>
#include "kone_device_scanner.h"
#include "kone.h"
#include "config.h"

static gint parameter_in_profile = -1;
static gint parameter_out_profile = -1;
static gchar *parameter_in_rmp = NULL;
static gchar *parameter_out_rmp = NULL;
static gint parameter_load = -1;
static gint parameter_save = -1;
static gboolean parameter_just_print_path = FALSE;
static gboolean parameter_read_firmware = FALSE;
static gboolean parameter_read_weight = FALSE;
static gboolean parameter_just_calibrate_tcu = FALSE;
static gboolean parameter_just_print_version = FALSE;
static gboolean parameter_just_print_actual_profile = FALSE;
static gboolean parameter_just_print_actual_dpi = FALSE;
static gint parameter_activate_profile = -1;

static GOptionEntry entries[] = {
		{ "read-profile", 'i', 0, G_OPTION_ARG_INT, &parameter_in_profile, "read profile NUMBER", "NUMBER" },
		{ "write-profile", 'o', 0, G_OPTION_ARG_INT, &parameter_out_profile, "write to profile NUMBER", "NUMBER" },
		{ "read-rmp", 'r', 0, G_OPTION_ARG_FILENAME, &parameter_in_rmp, "read from rmp-file FILENAME", "FILENAME" },
		{ "write-rmp", 'w', 0, G_OPTION_ARG_FILENAME, &parameter_out_rmp, "write to rmp-file FILENAME", "FILENAME" },
		{ "load", 'l', 0, G_OPTION_ARG_INT, &parameter_load, "load profile NUMBER", "NUMBER" },
		{ "save", 's', 0, G_OPTION_ARG_INT, &parameter_save, "save as profile NUMBER", "NUMBER" },
		{ "path", 'p', 0, G_OPTION_ARG_NONE, &parameter_just_print_path, "print path in sysfs", NULL },
		{ "read-firmware", 'f', 0, G_OPTION_ARG_NONE, &parameter_read_firmware, "read firmware version", NULL },
		{ "read-weight", 'g', 0, G_OPTION_ARG_NONE, &parameter_read_weight, "read fast-clip weight", NULL },
		{ "calibrate", 'c', 0, G_OPTION_ARG_NONE, &parameter_just_calibrate_tcu, "calibrate and activate tcu", NULL },
		{ "version", 'v', 0, G_OPTION_ARG_NONE, &parameter_just_print_version, "print version", NULL },
		{ "activate-profile", 'a', 0, G_OPTION_ARG_INT, &parameter_activate_profile, "activate profile NUMBER", "NUMBER" },
		{ "actual-profile", 0, 0, G_OPTION_ARG_NONE, &parameter_just_print_actual_profile, "print actual profile", NULL },
		{ "actual-dpi", 0, 0, G_OPTION_ARG_NONE, &parameter_just_print_actual_dpi, "print actual dpi", NULL },
		{ NULL }
};

static gboolean parse_profile_number_parameter(gint number, GError **error) {
	if (number < KONE_PROFILE_MIN || number > KONE_PROFILE_MAX) {
		g_set_error(error, G_OPTION_ERROR, G_OPTION_ERROR_BAD_VALUE, "profile numbers have to be in range of 1-5");
		return FALSE;
	}
	return TRUE;
}

static gboolean post_parse_func(GOptionContext *context, GOptionGroup *group, gpointer data, GError **error) {
	guint counter;

	counter = 0;
	if (parameter_just_print_path) ++counter;
	if (parameter_just_calibrate_tcu) ++counter;
	if (parameter_just_print_version) ++counter;
	if (parameter_just_print_actual_profile) ++counter;
	if (parameter_just_print_actual_dpi) ++counter;
	if (counter > 1) {
		g_set_error(error, G_OPTION_ERROR, G_OPTION_ERROR_FAILED, "just give one of -m, -e, -p, -c, -v, --actual-profile, --actual-dpi");
		return FALSE;
	}

	counter = 0;
	if (parameter_load != -1) ++counter;
	if (parameter_in_profile != -1) ++counter;
	if (parameter_in_rmp) ++counter;
	if (counter > 1) {
		g_set_error(error, G_OPTION_ERROR, G_OPTION_ERROR_FAILED, "-r, -l and -i are mutual exclusive");
		return FALSE;
	}

	if (parameter_activate_profile != -1 && !parse_profile_number_parameter(parameter_activate_profile, error))
		return FALSE;

	if (parameter_load != -1 && !parse_profile_number_parameter(parameter_load, error))
		return FALSE;

	if (parameter_save != -1 && !parse_profile_number_parameter(parameter_save, error))
		return FALSE;

	/*
	 * profile_load is allowed without writing afterwards, since warning about
	 * differing profiles is useful in itself.
	 */
	if ((parameter_in_profile != -1 || parameter_in_rmp || (parameter_load != -1)) &&
			!(parameter_out_profile != -1 || parameter_out_rmp || (parameter_save != -1))) {
		g_set_error(error, G_OPTION_ERROR, G_OPTION_ERROR_FAILED, "if -i or -r or -l is given, -o or -w or -s is needed");
		return FALSE;
	}

	return TRUE;
}

static GOptionContext *commandline_parse(int *argc, char ***argv) {
	GOptionContext *context = NULL;
	GError *error = NULL;

	context = g_option_context_new("- control extended capabilities of Roccat Kone mouse");
	g_option_context_add_main_entries(context, entries, NULL);

	g_option_group_set_parse_hooks(g_option_context_get_main_group(context), NULL, post_parse_func);

	if (!g_option_context_parse(context, argc, argv, &error)) {
		g_critical("option parsing failed: %s", error->message);
		exit(EXIT_FAILURE);
	}
	return context;
}

static void commandline_free(GOptionContext *context) {
	g_free(parameter_in_rmp);
	g_free(parameter_out_rmp);
	g_option_context_free(context);
}

static gboolean print_firmware(RoccatDevice *kone, GError **error) {
	gchar *firmware_string;
	guint firmware_version = kone_firmware_version_read(kone, error);
	if (*error)
		return FALSE;

	firmware_string = kone_firmware_version_to_string(firmware_version);
	g_print("%s\n", firmware_string);
	g_free(firmware_string);
	return TRUE;
}

static gboolean print_weight(RoccatDevice *kone, GError **error) {
	guint weight = kone_weight_read(kone, error);
	if (*error)
		return FALSE;
	g_print("%ig\n", weight);
	return TRUE;
}

static gboolean print_actual_dpi(RoccatDevice *kone, GError **error) {
	guint dpi = kone_actual_dpi_read(kone, error);
	if (*error)
		return FALSE;
	g_print("%i\n", dpi);
	return TRUE;
}

static gboolean print_actual_profile(RoccatDevice *kone, GError **error) {
	guint profile_number = kone_actual_profile_read(kone, error);
	if (*error)
		return FALSE;
	g_print("%i\n", profile_number);
	return TRUE;
}

static gboolean user_decision(gchar *text) {
	gchar decision;
	int retval;

	g_print("%s", text);
	g_print(" (y|n)");
	retval = scanf("%c", &decision);
	if (retval != 1)
		return FALSE;
	fflush(stdin);
	return g_ascii_tolower(decision) == 'y';
}

static gboolean save_profile(RoccatDevice *kone, KoneRMP *rmp, GError **error) {
	KoneProfile *profile;
	int retval;

	if (parameter_out_profile != -1)
		return FALSE;

	profile = kone_rmp_to_profile(rmp);
	retval = kone_profile_write(kone, parameter_out_profile, profile, error);
	g_free(profile);

	if (!retval)
		return FALSE;

	return TRUE;
}

static KoneRMP *get_profile(RoccatDevice *kone, GError **error) {
	KoneProfile *profile;
	KoneRMP *result;

	if (parameter_in_profile == -1)
		return NULL;

	profile = kone_profile_read(kone, parameter_in_profile, error);
	if (*error)
		return NULL;

	result = kone_rmp_dup(kone_default_rmp());

	kone_rmp_update_with_profile(result, profile);
	g_free(profile);

	return result;
}

static gboolean calibrate(RoccatDevice *kone, GError **error) {
	if (user_decision("To calibrate, you have to keep your hands off your Kone for 5 seconds."
			" Do you want to continue?")) {
		g_print("Please wait...");
		if (!kone_calibrate(kone, error))
			return FALSE;

		g_print("Done.\n");
	} else
		g_print("Calibration was aborted on user request.\n");
	return TRUE;
}

int main(int argc, char **argv) {
	GOptionContext *context;
	RoccatDevice *kone;
	GError *error = NULL;
	KoneRMP *rmp = NULL;
	int retval = EXIT_SUCCESS;

	roccat_secure();

	context = commandline_parse(&argc, &argv);

	if (parameter_just_print_version) {
		g_print(VERSION_STRING "\n");
		goto exit1;
	}

#if !(GLIB_CHECK_VERSION(2, 36, 0))
	g_type_init();
#endif

	kone = kone_device_first();
	if (kone == NULL) {
		g_critical("No Kone found.");
		retval = EXIT_FAILURE;
		goto exit1;
	}

	if (parameter_just_print_path) {
		g_print("%s", roccat_device_get_syspath(kone));
		goto exit2;
	}

	if (parameter_just_print_actual_profile) {
		if (!print_actual_profile(kone, &error)) {
			g_critical("print actual profile failed: %s", error->message);
			retval = EXIT_FAILURE;
		}
		goto exit2;
	}

	if (parameter_just_print_actual_dpi) {
		if (!print_actual_dpi(kone, &error)) {
			g_critical("print actual dpi failed: %s", error->message);
			retval = EXIT_FAILURE;
		}
		goto exit2;
	}

	if (parameter_just_calibrate_tcu) {
		if (!calibrate(kone, &error)) {
			g_critical("Error calibrating: %s", error->message);
			retval = EXIT_FAILURE;
		}
		goto exit2;
	}

	if (parameter_read_firmware) {
		if (!print_firmware(kone, &error)) {
			g_critical("Print firmware version failed: %s", error->message);
			retval = EXIT_FAILURE;
			goto exit2;
		}
	}

	if (parameter_read_weight) {
		if (!print_weight(kone, &error)) {
			g_critical("print weight failed: %s", error->message);
			retval = EXIT_FAILURE;
			goto exit2;
		}
	}

	if (parameter_load != -1) {
		rmp = kone_rmp_load(kone, parameter_load, &error);
		if (!rmp) {
			g_critical("Error loading profile %i: %s", parameter_load, error->message);
			retval = EXIT_FAILURE;
			goto exit2;
		}
	} else if (parameter_in_profile != -1) {
		rmp = get_profile(kone, &error);
		if (!rmp) {
			g_critical("Error reading profile: %s", error->message);
			retval = EXIT_FAILURE;
			goto exit2;
		}
	} else if (parameter_in_rmp) {
		rmp = kone_rmp_read_with_path(parameter_in_rmp, &error);
		if (!rmp) {
			g_critical("Error reading rmp: %s", error->message);
			retval = EXIT_FAILURE;
			goto exit2;
		}
	} else
		rmp = kone_rmp_dup(kone_default_rmp());

	kone_rmp_set_modified(rmp);

	if (parameter_save != -1) {
		if (!kone_rmp_save(kone, rmp, parameter_save, &error)) {
			g_critical("Error saving profile %i: %s", parameter_save, error->message);
			retval = EXIT_FAILURE;
			goto exit3;
		}
		kone_dbus_emit_profile_data_changed_outside_instant(parameter_save);
	}

	if (parameter_out_profile != -1) {
		if (!save_profile(kone, rmp, &error)) {
			g_critical("Error writing profile: %s", error->message);
			retval = EXIT_FAILURE;
			goto exit3;
		}
	}

	if (parameter_out_rmp) {
		if (!kone_rmp_write_with_path(parameter_out_rmp, rmp, &error)) {
			g_critical("Error writing rmp: %s", error->message);
			retval = EXIT_FAILURE;
			goto exit3;
		}
	}

	if (parameter_activate_profile != -1) {
		if (!kone_profile_activate(kone, parameter_activate_profile, &error)) {
			g_critical("Error activating profile: %s", error->message);
			retval = EXIT_FAILURE;
			goto exit3;
		}
		kone_dbus_emit_profile_changed_outside_instant(parameter_activate_profile);
	}

exit3:
	kone_rmp_free(rmp);
exit2:
	g_object_unref(G_OBJECT(kone));
exit1:
	commandline_free(context);
	g_clear_error(&error);
	exit(retval);
}
