/*
 * This file is part of roccat-tools.
 *
 * roccat-tools is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 *
 * roccat-tools is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with roccat-tools. If not, see <http://www.gnu.org/licenses/>.
 */

#include "konepuremilitary_eventhandler.h"
#include "konepuremilitary_eventhandler_channel.h"
#include "konepuremilitary_dbus_server.h"
#include "roccat_notification_cpi.h"
#include "roccat_notification_profile.h"
#include "roccat_notification_sensitivity.h"
#include "roccat_notification_timer.h"
#include "roccat_process_helper.h"
#include "roccat_talk.h"
#include "konepuremilitary.h"
#include "config.h"
#include <glib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>

#define KONEPUREMILITARY_EVENTHANDLER_CLASS(klass) (G_TYPE_CHECK_CLASS_CAST((klass), KONEPUREMILITARY_EVENTHANDLER_TYPE, KonepuremilitaryEventhandlerClass))
#define IS_KONEPUREMILITARY_EVENTHANDLER_CLASS(klass) (G_TYPE_CHECK_CLASS_TYPE((klass), KONEPUREMILITARY_EVENTHANDLER_TYPE))
#define KONEPUREMILITARY_EVENTHANDLER_GET_PRIVATE(obj) (G_TYPE_INSTANCE_GET_PRIVATE((obj), KONEPUREMILITARY_EVENTHANDLER_TYPE, KonepuremilitaryEventhandlerPrivate))

typedef struct _KonepuremilitaryEventhandlerClass KonepuremilitaryEventhandlerClass;

struct _KonepuremilitaryEventhandlerClass {
	GObjectClass parent_class;
};

struct _KonepuremilitaryEventhandlerPrivate {
	RoccatEventhandlerHost *host;
	KonepuremilitaryDBusServer *dbus_server;
	RoccatDeviceScanner *device_scanner;
	RoccatDevice *device;
	gboolean device_set_up;
	RoccatKeyFile *config;
	guint actual_profile_index;
	guint actual_sensitivity_x;
	KoneplusRmp *rmp[KONEPLUS_PROFILE_NUM];
	KonepuremilitaryEventhandlerChannel *channel;
	gulong active_window_changed_handler;
	RoccatNotificationProfile *profile_note;
	RoccatNotificationCpi *cpi_note;
	RoccatNotificationSensitivity *sensitivity_note;
	RoccatNotificationTimer *timer_note;

	guint8 active_talk_type;
	guint16 active_talk_device;
};

enum {
	PROP_0,
	PROP_HOST,
};

static void profile_changed(KonepuremilitaryEventhandler *eventhandler, guint profile_number) {
	KonepuremilitaryEventhandlerPrivate *priv = eventhandler->priv;
	guint profile_index = profile_number - 1;

	priv->actual_profile_index = profile_index;

	roccat_notification_profile_update(priv->profile_note, koneplus_configuration_get_profile_notification_type(priv->config),
			koneplus_configuration_get_notification_volume(priv->config),
			profile_number, koneplus_rmp_get_profile_name(priv->rmp[profile_index]));

	konepuremilitary_dbus_server_emit_profile_changed(priv->dbus_server, profile_number);
}

static gboolean set_profile(KonepuremilitaryEventhandler *eventhandler, guint profile_number) {
	KonepuremilitaryEventhandlerPrivate *priv = eventhandler->priv;
	guint profile_index;
	GError *local_error = NULL;

	if (profile_number == 0)
		return TRUE;

	profile_index = profile_number - 1;

	if (priv->actual_profile_index != profile_index) {
		if (!koneplus_actual_profile_write(priv->device, profile_index, &local_error)) {
			g_warning("Couldn't set profile %u: %s", profile_number, local_error->message);
			g_clear_error(&local_error);
			return FALSE;
		}
		profile_changed(eventhandler, profile_number);
	}
	return TRUE;
}

static void window_changed_cb(RoccatEventhandlerHost *host, gchar const *title, gpointer user_data) {
	KonepuremilitaryEventhandler *eventhandler = KONEPUREMILITARY_EVENTHANDLER(user_data);
	KonepuremilitaryEventhandlerPrivate *priv = eventhandler->priv;
	guint profile_index;
	guint game_file_index;
	gchar *pattern;

	for (profile_index = 0; profile_index < KONEPLUS_PROFILE_NUM; ++profile_index) {

		if (priv->rmp[profile_index] == NULL)
			continue;

		for (game_file_index = 0; game_file_index < KONEPLUS_GAMEFILE_NUM; ++game_file_index) {
			pattern = koneplus_rmp_get_game_file_name(priv->rmp[profile_index], game_file_index);
			if (strcmp(pattern, "") == 0) {
				g_free(pattern);
				continue;
			}

			if (g_regex_match_simple(pattern, title, 0, 0)) {
				set_profile(eventhandler, profile_index + 1);
				g_free(pattern);
				return;
			}
			g_free(pattern);
		}
	}

	set_profile(eventhandler, koneplus_configuration_get_default_profile_number(priv->config));
	return;
}

static gboolean talk_easyshift(KonepuremilitaryEventhandler *eventhandler, gboolean state) {
	KonepuremilitaryEventhandlerPrivate *priv = eventhandler->priv;
	GError *error = NULL;
	gboolean retval = FALSE;

	if (priv->device_set_up) {
		retval = koneplus_talk_easyshift(priv->device,
				(state != FALSE) ? KONEPLUS_TALK_EASYSHIFT_ON : KONEPLUS_TALK_EASYSHIFT_OFF,
				&error);
		if (error) {
			g_warning("Error issuing easyshift: %s", error->message);
			g_clear_error(&error);
		}
	}

	return retval;
}

static gboolean talk_easyshift_lock(KonepuremilitaryEventhandler *eventhandler, gboolean state) {
	KonepuremilitaryEventhandlerPrivate *priv = eventhandler->priv;
	GError *error = NULL;
	gboolean retval = FALSE;

	if (priv->device_set_up) {
		retval = koneplus_talk_easyshift_lock(priv->device,
				(state != FALSE) ? KONEPLUS_TALK_EASYSHIFT_ON : KONEPLUS_TALK_EASYSHIFT_OFF,
				&error);
		if (error) {
			g_warning("Error issuing easyshift_lock: %s", error->message);
			g_clear_error(&error);
		}
	}

	return retval;
}

static gboolean talk_easyaim(KonepuremilitaryEventhandler *eventhandler, guchar state) {
	KonepuremilitaryEventhandlerPrivate *priv = eventhandler->priv;
	GError *error = NULL;
	gboolean retval = FALSE;

	if (priv->device_set_up) {
		retval = koneplus_talk_easyaim(priv->device, state, &error);
		if (error) {
			g_warning("Error issuing easyaim: %s", error->message);
			g_clear_error(&error);
		}
	}

	return retval;
}

static gboolean set_led_rgb(KonepuremilitaryEventhandler *eventhandler, guint32 effect, guint32 ambient_color, guint32 event_color) {
	KonepuremilitaryEventhandlerPrivate *priv = eventhandler->priv;
	GError *error = NULL;
	gboolean retval = FALSE;

	if (priv->device_set_up) {
		retval = koneplus_talkfx(priv->device, effect, ambient_color, event_color, &error);
		if (error) {
			g_warning("Error issuing talkfx: %s", error->message);
			g_clear_error(&error);
		}
	}

	return retval;
}

static gboolean restore_led_rgb(KonepuremilitaryEventhandler *eventhandler) {
	KonepuremilitaryEventhandlerPrivate *priv = eventhandler->priv;
	GError *error = NULL;
	gboolean retval = FALSE;

	if (priv->device_set_up) {
		retval = koneplus_talkfx_off(priv->device, &error);
		if (error) {
			g_warning("Error issuing talkfx_off: %s", error->message);
			g_clear_error(&error);
		}
	}

	return retval;
}

static guint konepuremilitary_eventhandler_get_talk_device(RoccatEventhandlerPlugin *self) {
	KonepuremilitaryEventhandlerPrivate *priv = KONEPUREMILITARY_EVENTHANDLER(self)->priv;

	if (!priv->device_set_up)
		return 0;

	return gaminggear_device_get_product_id(GAMINGGEAR_DEVICE(priv->device));
}

static void konepuremilitary_eventhandler_talk_easyshift(RoccatEventhandlerPlugin *self, gboolean state) {
	KonepuremilitaryEventhandler *eventhandler = KONEPUREMILITARY_EVENTHANDLER(self);
	talk_easyshift(eventhandler, state);
}

static void konepuremilitary_eventhandler_talk_easyshift_lock(RoccatEventhandlerPlugin *self, gboolean state) {
	KonepuremilitaryEventhandler *eventhandler = KONEPUREMILITARY_EVENTHANDLER(self);
	talk_easyshift_lock(eventhandler, state);
}

static void konepuremilitary_eventhandler_talk_easyaim(RoccatEventhandlerPlugin *self, guchar state) {
	KonepuremilitaryEventhandler *eventhandler = KONEPUREMILITARY_EVENTHANDLER(self);
	talk_easyaim(eventhandler, state);
}

static void konepuremilitary_eventhandler_set_led_rgb(RoccatEventhandlerPlugin *self, guint32 effect, guint32 ambient_color, guint32 event_color) {
	KonepuremilitaryEventhandler *eventhandler = KONEPUREMILITARY_EVENTHANDLER(self);
	set_led_rgb(eventhandler, effect, ambient_color, event_color);
}

static void konepuremilitary_eventhandler_restore_led_rgb(RoccatEventhandlerPlugin *self) {
	KonepuremilitaryEventhandler *eventhandler = KONEPUREMILITARY_EVENTHANDLER(self);
	restore_led_rgb(eventhandler);
}

static void konepuremilitary_roccat_eventhandler_interface_init(RoccatEventhandlerPluginInterface *iface) {
	iface->get_talk_device = konepuremilitary_eventhandler_get_talk_device;
	iface->talk_easyshift = konepuremilitary_eventhandler_talk_easyshift;
	iface->talk_easyshift_lock = konepuremilitary_eventhandler_talk_easyshift_lock;
	iface->talk_easyaim = konepuremilitary_eventhandler_talk_easyaim;
	iface->set_led_rgb = konepuremilitary_eventhandler_set_led_rgb;
	iface->restore_led_rgb = konepuremilitary_eventhandler_restore_led_rgb;
}

G_DEFINE_TYPE_WITH_CODE(KonepuremilitaryEventhandler, konepuremilitary_eventhandler, G_TYPE_OBJECT,
		G_IMPLEMENT_INTERFACE(ROCCAT_EVENTHANDLER_PLUGIN_TYPE, konepuremilitary_roccat_eventhandler_interface_init));

GObject *konepuremilitary_eventhandler_new(RoccatEventhandlerHost *host) {
	return g_object_new(KONEPUREMILITARY_EVENTHANDLER_TYPE,
			"host", host,
			NULL);
}

static void talk_easyshift_cb(KonepuremilitaryDBusServer *server, guchar state, gpointer user_data) {
	KonepuremilitaryEventhandler *eventhandler = KONEPUREMILITARY_EVENTHANDLER(user_data);
	talk_easyshift(eventhandler, state);
}

static void talk_easyshift_lock_cb(KonepuremilitaryDBusServer *server, guchar state, gpointer user_data) {
	KonepuremilitaryEventhandler *eventhandler = KONEPUREMILITARY_EVENTHANDLER(user_data);
	talk_easyshift_lock(eventhandler, state);
}

static void talk_easyaim_cb(KonepuremilitaryDBusServer *server, guchar state, gpointer user_data) {
	KonepuremilitaryEventhandler *eventhandler = KONEPUREMILITARY_EVENTHANDLER(user_data);
	talk_easyaim(eventhandler, state);
}

static void set_led_rgb_cb(KonepuremilitaryDBusServer *server, guint effect, guint ambient_color, guint event_color, gpointer user_data) {
	KonepuremilitaryEventhandler *eventhandler = KONEPUREMILITARY_EVENTHANDLER(user_data);
	set_led_rgb(eventhandler, effect, ambient_color, event_color);
}

static void restore_led_rgb_cb(KonepuremilitaryDBusServer *server, gpointer user_data) {
	KonepuremilitaryEventhandler *eventhandler = KONEPUREMILITARY_EVENTHANDLER(user_data);
	restore_led_rgb(eventhandler);
}

static void open_gui_cb(gpointer source, gpointer user_data) {
	gchar *path;
	path = g_build_path("/", INSTALL_PREFIX, "bin", "konepuremilitaryconfig", (gchar const *)NULL);
	roccat_double_fork(path);
	g_free(path);
}

static void profile_data_changed_outside_cb(KonepuremilitaryDBusServer *server, guchar number, gpointer user_data) {
	KonepuremilitaryEventhandler *eventhandler = KONEPUREMILITARY_EVENTHANDLER(user_data);
	KonepuremilitaryEventhandlerPrivate *priv = eventhandler->priv;
	guint profile_index = number - 1;

	/* Just load rmp without updating data from hardware */
	koneplus_rmp_free(priv->rmp[profile_index]);
	priv->rmp[profile_index] = konepuremilitary_rmp_load_actual(profile_index);
}

static void configuration_reload(KonepuremilitaryEventhandler *eventhandler) {
	KonepuremilitaryEventhandlerPrivate *priv = eventhandler->priv;
	if (priv->config)
		koneplus_configuration_free(priv->config);
	priv->config = konepuremilitary_configuration_load();
}

static void configuration_changed_outside_cb(KonepuremilitaryDBusServer *server, gpointer user_data) {
	KonepuremilitaryEventhandler *eventhandler = KONEPUREMILITARY_EVENTHANDLER(user_data);
	configuration_reload(eventhandler);
}

static void cpi_changed_cb(KonepuremilitaryEventhandlerChannel *channel, guchar cpi_index, gpointer user_data) {
	KonepuremilitaryEventhandler *eventhandler = KONEPUREMILITARY_EVENTHANDLER(user_data);
	KonepuremilitaryEventhandlerPrivate *priv = eventhandler->priv;

	roccat_notification_cpi_update(priv->cpi_note,
			koneplus_configuration_get_cpi_notification_type(priv->config),
			koneplus_configuration_get_notification_volume(priv->config),
			koneplus_rmp_get_cpi_level_x(priv->rmp[priv->actual_profile_index], cpi_index) * 100,
			koneplus_rmp_get_cpi_level_x(priv->rmp[priv->actual_profile_index], cpi_index) * 100);
}

static void profile_changed_cb(gpointer source, guchar profile_number, gpointer user_data) {
	KonepuremilitaryEventhandler *eventhandler = KONEPUREMILITARY_EVENTHANDLER(user_data);
	profile_changed(eventhandler, profile_number);
}

static void quicklaunch_cb(KonepuremilitaryEventhandlerChannel *channel, guchar key_index, gpointer user_data) {
	KonepuremilitaryEventhandler *eventhandler = KONEPUREMILITARY_EVENTHANDLER(user_data);
	KonepuremilitaryEventhandlerPrivate *priv = eventhandler->priv;
	KoneplusRmpMacroKeyInfo *macro_key_info;

	macro_key_info = koneplus_rmp_get_macro_key_info(priv->rmp[priv->actual_profile_index], key_index);
	koneplus_play_quicklaunch_forked(macro_key_info);
	koneplus_rmp_macro_key_info_free(macro_key_info);
}

static void sensitivity_changed_cb(KonepuremilitaryEventhandlerChannel *channel, guchar x, guchar y, gpointer user_data) {
	KonepuremilitaryEventhandler *eventhandler = KONEPUREMILITARY_EVENTHANDLER(user_data);
	KonepuremilitaryEventhandlerPrivate *priv = eventhandler->priv;
	gint notification_type;

	notification_type = koneplus_configuration_get_sensitivity_notification_type(priv->config);

	if (notification_type == ROCCAT_NOTIFICATION_TYPE_OSD)
		roccat_notification_sensitivity_update(priv->sensitivity_note, notification_type, x - 6, y - 6);
	else if (x > priv->actual_sensitivity_x)
		roccat_notification_sensitivity_update_up(priv->sensitivity_note, notification_type,
				koneplus_configuration_get_notification_volume(priv->config));
	else
		roccat_notification_sensitivity_update_down(priv->sensitivity_note, notification_type,
				koneplus_configuration_get_notification_volume(priv->config));

	priv->actual_sensitivity_x = x;
}

static void talk_cb(KonepuremilitaryEventhandlerChannel *channel, guchar key_index, guchar event, gpointer user_data) {
	KonepuremilitaryEventhandler *eventhandler = KONEPUREMILITARY_EVENTHANDLER(user_data);
	KonepuremilitaryEventhandlerPrivate *priv = eventhandler->priv;
	KoneplusRmpMacroKeyInfo *macro_key_info;
	static gboolean state = FALSE;

	if (event == KONEPLUS_CHRDEV_EVENT_ACTION_PRESS) {
		macro_key_info = koneplus_rmp_get_macro_key_info(priv->rmp[priv->actual_profile_index], key_index);
		priv->active_talk_type = macro_key_info->type;
		priv->active_talk_device = koneplus_rmp_macro_key_info_get_talk_device(macro_key_info);
		koneplus_rmp_macro_key_info_free(macro_key_info);
	}

	switch (priv->active_talk_type) {
	case KONEPLUS_PROFILE_BUTTON_TYPE_EASYSHIFT:
		break;
	case KONEPLUS_PROFILE_BUTTON_TYPE_TALK_BOTH_EASYSHIFT:
		roccat_eventhandler_plugin_emit_talk_easyshift(ROCCAT_EVENTHANDLER_PLUGIN(eventhandler),
				ROCCAT_TALK_DEVICE_KEYBOARD,
				(event == KONEPLUS_CHRDEV_EVENT_ACTION_PRESS) ? TRUE : FALSE);
		break;
	case KONEPLUS_PROFILE_BUTTON_TYPE_TALK_OTHER_EASYSHIFT:
		roccat_eventhandler_plugin_emit_talk_easyshift(ROCCAT_EVENTHANDLER_PLUGIN(eventhandler),
				priv->active_talk_device,
				(event == KONEPLUS_CHRDEV_EVENT_ACTION_PRESS) ? TRUE : FALSE);
		break;
	case KONEPLUS_PROFILE_BUTTON_TYPE_TALK_OTHER_EASYSHIFT_LOCK:
		if (event == KONEPLUS_CHRDEV_EVENT_ACTION_PRESS) {
			state = !state;
			roccat_eventhandler_plugin_emit_talk_easyshift_lock(ROCCAT_EVENTHANDLER_PLUGIN(eventhandler),
					priv->active_talk_device,
					state);
		}
		break;
	default:
		g_debug("unknown talk event of type 0x%02x", priv->active_talk_type);
	}

	if (event == KONEPLUS_CHRDEV_EVENT_ACTION_RELEASE)
		priv->active_talk_type = KONEPLUS_PROFILE_BUTTON_TYPE_DISABLED;
}

static void timer_start_cb(KonepuremilitaryEventhandlerChannel *channel, guchar key_index, gpointer user_data) {
	KonepuremilitaryEventhandler *eventhandler = KONEPUREMILITARY_EVENTHANDLER(user_data);
	KonepuremilitaryEventhandlerPrivate *priv = eventhandler->priv;
	KoneplusRmpMacroKeyInfo *macro_key_info;

	macro_key_info = koneplus_rmp_get_macro_key_info(priv->rmp[priv->actual_profile_index], key_index);
	roccat_notification_timer_start(priv->timer_note,
			koneplus_configuration_get_timer_notification_type(priv->config),
			koneplus_configuration_get_notification_volume(priv->config),
			(gchar const *)macro_key_info->timer_name,
			koneplus_rmp_macro_key_info_get_timer_length(macro_key_info));
	koneplus_rmp_macro_key_info_free(macro_key_info);
}

static void timer_stop_cb(KonepuremilitaryEventhandlerChannel *channel, gpointer user_data) {
	KonepuremilitaryEventhandler *eventhandler = KONEPUREMILITARY_EVENTHANDLER(user_data);
	KonepuremilitaryEventhandlerPrivate *priv = eventhandler->priv;

	roccat_notification_timer_stop(priv->timer_note);
}

static gboolean profile_data_load(KonepuremilitaryEventhandler *eventhandler, GError **error) {
	KonepuremilitaryEventhandlerPrivate *priv = eventhandler->priv;
	guint i;

	for (i = 0; i < KONEPLUS_PROFILE_NUM; ++i) {
		priv->rmp[i] = konepuremilitary_rmp_load(priv->device, i, error);
		if (priv->rmp[i] == NULL)
			return FALSE;
	}
	return TRUE;
}

static void device_add_cb(RoccatDeviceScannerInterface *interface, RoccatDevice *device, gpointer user_data) {
	KonepuremilitaryEventhandler *eventhandler = KONEPUREMILITARY_EVENTHANDLER(user_data);
	KonepuremilitaryEventhandlerPrivate *priv = eventhandler->priv;
	GError *error = NULL;
	KonepuremilitaryProfileSettings *profile_settings;

	if (priv->device) {
		g_critical("Multiple devices present.");
		return;
	}

	priv->device = device;
	g_object_ref(G_OBJECT(device));
	roccat_device_debug(device);

	configuration_reload(eventhandler);

	/* Notificator gets initialized only once if a device is found.
	 * Gets deinited once in finalizer.
	 */
	if (!priv->profile_note)
		priv->profile_note = roccat_notification_profile_new(roccat_eventhandler_host_get_notificator(priv->host), KONEPUREMILITARY_DEVICE_NAME);
	if (!priv->cpi_note)
		priv->cpi_note = roccat_notification_cpi_new(roccat_eventhandler_host_get_notificator(priv->host), KONEPUREMILITARY_DEVICE_NAME);
	if (!priv->sensitivity_note)
		priv->sensitivity_note = roccat_notification_sensitivity_new(roccat_eventhandler_host_get_notificator(priv->host), KONEPUREMILITARY_DEVICE_NAME);
	if (!priv->timer_note)
		priv->timer_note = roccat_notification_timer_new(roccat_eventhandler_host_get_notificator(priv->host), KONEPUREMILITARY_DEVICE_NAME);

	priv->actual_profile_index = koneplus_actual_profile_read(priv->device, &error);
	if (error) {
		g_warning("Error reading actual profile: %s", error->message);
		g_clear_error(&error);
		priv->actual_profile_index = 0; /* guessing */
	} else {
		profile_settings = konepuremilitary_profile_settings_read(priv->device, priv->actual_profile_index, &error);
		if (error) {
			g_warning("Error reading profile settings: %s", error->message);
			g_clear_error(&error);
			priv->actual_sensitivity_x = (KONEPLUS_PROFILE_SETTINGS_SENSITIVITY_MIN + KONEPLUS_PROFILE_SETTINGS_SENSITIVITY_MAX) >> 1; /* guessing */
		}
		else {
			priv->actual_sensitivity_x = profile_settings->sensitivity_x;
			g_free(profile_settings);
		}
	}

	profile_data_load(eventhandler, &error);
	if (error) {
		g_warning("Error reading profile data: %s", error->message);
		g_clear_error(&error);
	}

	priv->dbus_server = konepuremilitary_dbus_server_new();
	konepuremilitary_dbus_server_connect(priv->dbus_server);
	g_signal_connect(G_OBJECT(priv->dbus_server), "easyshift", G_CALLBACK(talk_easyshift_cb), eventhandler);
	g_signal_connect(G_OBJECT(priv->dbus_server), "easyshift-lock", G_CALLBACK(talk_easyshift_lock_cb), eventhandler);
	g_signal_connect(G_OBJECT(priv->dbus_server), "easyaim", G_CALLBACK(talk_easyaim_cb), eventhandler);
	g_signal_connect(G_OBJECT(priv->dbus_server), "set-led-rgb", G_CALLBACK(set_led_rgb_cb), eventhandler);
	g_signal_connect(G_OBJECT(priv->dbus_server), "restore-led-rgb", G_CALLBACK(restore_led_rgb_cb), eventhandler);
	g_signal_connect(G_OBJECT(priv->dbus_server), "open-gui", G_CALLBACK(open_gui_cb), eventhandler);
	g_signal_connect(G_OBJECT(priv->dbus_server), "profile-changed-outside", G_CALLBACK(profile_changed_cb), eventhandler);
	g_signal_connect(G_OBJECT(priv->dbus_server), "profile-data-changed-outside", G_CALLBACK(profile_data_changed_outside_cb), eventhandler);
	g_signal_connect(G_OBJECT(priv->dbus_server), "configuration-changed-outside", G_CALLBACK(configuration_changed_outside_cb), eventhandler);

	priv->active_window_changed_handler = g_signal_connect(G_OBJECT(priv->host), "active-window-changed", G_CALLBACK(window_changed_cb), eventhandler);

	konepuremilitary_eventhandler_channel_start(priv->channel, roccat_device_get_hidraw_endpoint_dev(priv->device, KONEPLUS_INTERFACE_MOUSE), &error);
	if (error) {
		g_warning("Error starting hidraw eventhandler: %s", error->message);
		g_clear_error(&error);
	}

	priv->active_talk_type = KONEPLUS_PROFILE_BUTTON_TYPE_DISABLED;

	priv->device_set_up = TRUE;
}

static void stop(KonepuremilitaryEventhandler *eventhandler) {
	KonepuremilitaryEventhandlerPrivate *priv = eventhandler->priv;
	guint i;

	priv->device_set_up = FALSE;

	g_object_unref(priv->dbus_server);
	priv->dbus_server = NULL;

	konepuremilitary_eventhandler_channel_stop(priv->channel);

	g_signal_handler_disconnect(G_OBJECT(priv->host), priv->active_window_changed_handler);

	if (priv->device) {
		g_object_unref(G_OBJECT(priv->device));
		priv->device = NULL;
	}

	koneplus_configuration_free(priv->config);
	priv->config = NULL;

	for (i = 0; i < KONEPLUS_PROFILE_NUM; ++i) {
		koneplus_rmp_free(priv->rmp[i]);
		priv->rmp[i] = NULL;
	}
}

static void device_remove_cb(RoccatDeviceScanner *scanner, RoccatDevice *device, gpointer user_data) {
	KonepuremilitaryEventhandler *eventhandler = KONEPUREMILITARY_EVENTHANDLER(user_data);
	KonepuremilitaryEventhandlerPrivate *priv = eventhandler->priv;
	if (priv->device && gaminggear_device_equal(GAMINGGEAR_DEVICE(device), GAMINGGEAR_DEVICE(priv->device)))
		stop(eventhandler);
}

static void konepuremilitary_eventhandler_init(KonepuremilitaryEventhandler *eventhandler) {
	eventhandler->priv = KONEPUREMILITARY_EVENTHANDLER_GET_PRIVATE(eventhandler);
}

static void finalize(GObject *object) {
	KonepuremilitaryEventhandler *eventhandler = KONEPUREMILITARY_EVENTHANDLER(object);
	KonepuremilitaryEventhandlerPrivate *priv = eventhandler->priv;

	g_object_unref(priv->channel);
	priv->channel = NULL;

	if (priv->device_scanner) {
		g_object_unref(G_OBJECT(priv->device_scanner));
		priv->device_scanner = NULL;
	}

	stop(eventhandler);

	if (priv->timer_note) {
		roccat_notification_timer_free(priv->timer_note);
		priv->timer_note = NULL;
	}

	if (priv->profile_note) {
		roccat_notification_profile_free(priv->profile_note);
		priv->profile_note = NULL;
	}

	if (priv->cpi_note) {
		roccat_notification_cpi_free(priv->cpi_note);
		priv->cpi_note = NULL;
	}

	if (priv->sensitivity_note) {
		roccat_notification_sensitivity_free(priv->sensitivity_note);
		priv->sensitivity_note = NULL;
	}

	G_OBJECT_CLASS(konepuremilitary_eventhandler_parent_class)->finalize(object);
}

static GObject *constructor(GType gtype, guint n_properties, GObjectConstructParam *properties) {
	KonepuremilitaryEventhandler *eventhandler;
	KonepuremilitaryEventhandlerPrivate *priv;
	GObject *obj;

	obj = G_OBJECT_CLASS(konepuremilitary_eventhandler_parent_class)->constructor(gtype, n_properties, properties);
	eventhandler = KONEPUREMILITARY_EVENTHANDLER(obj);
	priv = eventhandler->priv;

	priv->channel = konepuremilitary_eventhandler_channel_new();
	g_signal_connect(G_OBJECT(priv->channel), "cpi-changed", G_CALLBACK(cpi_changed_cb), eventhandler);
	g_signal_connect(G_OBJECT(priv->channel), "open-driver", G_CALLBACK(open_gui_cb), eventhandler);
	g_signal_connect(G_OBJECT(priv->channel), "profile-changed", G_CALLBACK(profile_changed_cb), eventhandler);
	g_signal_connect(G_OBJECT(priv->channel), "quicklaunch", G_CALLBACK(quicklaunch_cb), eventhandler);
	g_signal_connect(G_OBJECT(priv->channel), "sensitivity-changed", G_CALLBACK(sensitivity_changed_cb), eventhandler);
	g_signal_connect(G_OBJECT(priv->channel), "talk", G_CALLBACK(talk_cb), eventhandler);
	g_signal_connect(G_OBJECT(priv->channel), "timer-start", G_CALLBACK(timer_start_cb), eventhandler);
	g_signal_connect(G_OBJECT(priv->channel), "timer-stop", G_CALLBACK(timer_stop_cb), eventhandler);

	priv->device_scanner = konepuremilitary_device_scanner_new();
	g_signal_connect(G_OBJECT(priv->device_scanner), "device-added", G_CALLBACK(device_add_cb), eventhandler);
	g_signal_connect(G_OBJECT(priv->device_scanner), "device-removed", G_CALLBACK(device_remove_cb), eventhandler);
	roccat_device_scanner_interface_start(ROCCAT_DEVICE_SCANNER_INTERFACE(priv->device_scanner));

	return obj;
}

static void set_property(GObject *object, guint prop_id, GValue const *value, GParamSpec *pspec) {
	KonepuremilitaryEventhandlerPrivate *priv = KONEPUREMILITARY_EVENTHANDLER(object)->priv;

	switch(prop_id) {
	case PROP_HOST:
		priv->host = g_value_get_pointer(value);
		break;
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
	}
}

static void konepuremilitary_eventhandler_class_init(KonepuremilitaryEventhandlerClass *klass) {
	GObjectClass *gobject_class;

	gobject_class = G_OBJECT_CLASS(klass);

	gobject_class->constructor = constructor;
	gobject_class->finalize = finalize;
	gobject_class->set_property = set_property;

	g_type_class_add_private(klass, sizeof(KonepuremilitaryEventhandlerPrivate));

	g_object_class_install_property(gobject_class, PROP_HOST,
			g_param_spec_pointer("host",
					"Host",
					"Pointer to RoccatEventhandlerHost interface",
					G_PARAM_WRITABLE | G_PARAM_CONSTRUCT_ONLY));
}

ROCCAT_EVENTHANDLER_PLUGIN_API RoccatEventhandlerPlugin *roccat_eventhandler_plugin_new(RoccatEventhandlerHost *host) {
	return ROCCAT_EVENTHANDLER_PLUGIN(konepuremilitary_eventhandler_new(host));
}
