/*
 * This file is part of roccat-tools.
 *
 * roccat-tools is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 *
 * roccat-tools is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with roccat-tools. If not, see <http://www.gnu.org/licenses/>.
 */

#include <gaminggear/macro_threads.h>
#include "pyra_eventhandler.h"
#include "pyra_eventhandler_channel.h"
#include "pyra_dbus_server.h"
#include "roccat_notification_cpi.h"
#include "roccat_notification_profile.h"
#include "roccat_process_helper.h"
#include "pyra.h"
#include "config.h"
#include <glib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>

#define PYRA_EVENTHANDLER_CLASS(klass) (G_TYPE_CHECK_CLASS_CAST((klass), PYRA_EVENTHANDLER_TYPE, PyraEventhandlerClass))
#define IS_PYRA_EVENTHANDLER_CLASS(klass) (G_TYPE_CHECK_CLASS_TYPE((klass), PYRA_EVENTHANDLER_TYPE))
#define PYRA_EVENTHANDLER_GET_PRIVATE(obj) (G_TYPE_INSTANCE_GET_PRIVATE((obj), PYRA_EVENTHANDLER_TYPE, PyraEventhandlerPrivate))

typedef struct _PyraEventhandlerClass PyraEventhandlerClass;

struct _PyraEventhandlerClass {
	GObjectClass parent_class;
};

struct _PyraEventhandlerPrivate {
	RoccatEventhandlerHost *host;
	PyraDBusServer *dbus_server;
	RoccatDeviceScanner *device_scanner;
	RoccatDevice *device;
	gboolean device_set_up;
	RoccatKeyFile *config;
	guint actual_profile_index;
	PyraRMP *rmp[PYRA_PROFILE_NUM];
	PyraEventhandlerChannel *channel;
	gulong active_window_changed_handler;
	
	/*
	 * FIXME notificator für Fehlermeldungen verwenden,
	 * notificator refaktorisieren und Basis für beides verwenden oder
	 * Eigene Implementierung eines Fehlernotificators
	 */
	RoccatNotificationProfile *profile_note;
	RoccatNotificationCpi *cpi_note;

	int kbd_dev;
	int mouse_dev;
	GaminggearMacroThreads *threads;
};

enum {
	PROP_0,
	PROP_HOST,
};

static void profile_changed(PyraEventhandler *eventhandler, guint profile_number) {
	PyraEventhandlerPrivate *priv = eventhandler->priv;
	guint profile_index = profile_number - 1;

	priv->actual_profile_index = profile_index;

	roccat_notification_profile_update(priv->profile_note, pyra_configuration_get_profile_notification_type(priv->config),
			pyra_configuration_get_notification_volume(priv->config),
			profile_number, pyra_rmp_get_profile_name(priv->rmp[profile_index]));

	pyra_dbus_server_emit_profile_changed(priv->dbus_server, profile_number);
}

static gboolean set_profile(PyraEventhandler *eventhandler, guint profile_number) {
	PyraEventhandlerPrivate *priv = eventhandler->priv;
	guint profile_index;
	GError *local_error = NULL;

	if (profile_number == 0)
		return TRUE;

	profile_index = profile_number - 1;

	if (priv->actual_profile_index != profile_index) {
		if (!pyra_actual_profile_write(priv->device, profile_index, &local_error)) {
			g_warning("Couldn't set profile %u: %s", profile_number, local_error->message);
			g_clear_error(&local_error);
			return FALSE;
		}
		profile_changed(eventhandler, profile_number);
	}
	return TRUE;
}

static void window_changed_cb(RoccatEventhandlerHost *host, gchar const *title, gpointer user_data) {
	PyraEventhandler *eventhandler = PYRA_EVENTHANDLER(user_data);
	PyraEventhandlerPrivate *priv = eventhandler->priv;
	guint profile_index;
	guint game_file_index;
	gchar *pattern;

	for (profile_index = 0; profile_index < PYRA_PROFILE_NUM; ++profile_index) {

		if (priv->rmp[profile_index] == NULL)
			continue;

		for (game_file_index = 0; game_file_index < PYRA_GAMEFILE_NUM; ++game_file_index) {
			pattern = pyra_rmp_get_game_file_name(priv->rmp[profile_index], game_file_index);
			if (strcmp(pattern, "") == 0) {
				g_free(pattern);
				continue;
			}

			if (g_regex_match_simple(pattern, title, 0, 0)) {
				set_profile(eventhandler, profile_index + 1);
				g_free(pattern);
				return;
			}
			g_free(pattern);
		}
	}

	set_profile(eventhandler, pyra_configuration_get_default_profile_number(priv->config));
	return;
}

static guint pyra_eventhandler_get_talk_device(RoccatEventhandlerPlugin *self) {
	return 0;
}

static void pyra_roccat_eventhandler_interface_init(RoccatEventhandlerPluginInterface *iface) {
	iface->get_talk_device = pyra_eventhandler_get_talk_device;
	iface->talk_easyshift = NULL;
	iface->talk_easyshift_lock = NULL;
	iface->talk_easyaim = NULL;
	iface->set_led_rgb = NULL;
	iface->restore_led_rgb = NULL;
}

G_DEFINE_TYPE_WITH_CODE(PyraEventhandler, pyra_eventhandler, G_TYPE_OBJECT,
		G_IMPLEMENT_INTERFACE(ROCCAT_EVENTHANDLER_PLUGIN_TYPE, pyra_roccat_eventhandler_interface_init));

GObject *pyra_eventhandler_new(RoccatEventhandlerHost *host) {
	return g_object_new(PYRA_EVENTHANDLER_TYPE,
			"host", host,
			NULL);
}

static void open_gui_cb(gpointer source, gpointer user_data) {
	gchar *path;
	path = g_build_path("/", INSTALL_PREFIX, "bin", "pyraconfig", (gchar const *)NULL);
	roccat_double_fork(path);
	g_free(path);
}

static void profile_data_changed_outside_cb(PyraDBusServer *server, guchar number, gpointer user_data) {
	PyraEventhandler *eventhandler = PYRA_EVENTHANDLER(user_data);
	PyraEventhandlerPrivate *priv = eventhandler->priv;
	guint profile_index = number - 1;

	/* Just load rmp without updating data from hardware */
	pyra_rmp_free(priv->rmp[profile_index]);
	priv->rmp[profile_index] = pyra_rmp_load_actual(profile_index);
}

static void configuration_reload(PyraEventhandler *eventhandler) {
	PyraEventhandlerPrivate *priv = eventhandler->priv;
	if (priv->config)
		pyra_configuration_free(priv->config);
	priv->config = pyra_configuration_load();
}

static void configuration_changed_outside_cb(PyraDBusServer *server, gpointer user_data) {
	PyraEventhandler *eventhandler = PYRA_EVENTHANDLER(user_data);
	configuration_reload(eventhandler);
}

static void cpi_changed_cb(PyraEventhandlerChannel *channel, guchar cpi_raw, gpointer user_data) {
	PyraEventhandler *eventhandler = PYRA_EVENTHANDLER(user_data);
	PyraEventhandlerPrivate *priv = eventhandler->priv;
	guint cpi_value;
	
	cpi_value = pyra_cpi_raw_to_cpi(cpi_raw);
	roccat_notification_cpi_update(priv->cpi_note,
			pyra_configuration_get_cpi_notification_type(priv->config),
			pyra_configuration_get_notification_volume(priv->config),
			cpi_value, cpi_value);
}

static void profile_changed_cb(gpointer source, guchar profile_number, gpointer user_data) {
	PyraEventhandler *eventhandler = PYRA_EVENTHANDLER(user_data);
	profile_changed(eventhandler, profile_number);
}

static void macro_cb(PyraEventhandlerChannel *channel, guchar key_number, gpointer user_data) {
	PyraEventhandler *eventhandler = PYRA_EVENTHANDLER(user_data);
	PyraEventhandlerPrivate *priv = eventhandler->priv;
	PyraRMPMacroKeyInfo *macro_key_info;
	GaminggearMacro *gaminggear_macro;
	
	macro_key_info = pyra_rmp_get_macro_key_info(priv->rmp[priv->actual_profile_index], key_number - 1);
	gaminggear_macro = pyra_rmp_macro_key_info_to_gaminggear_macro(macro_key_info);
	g_free(macro_key_info);
	gaminggear_play_macro_threaded(priv->threads, key_number - 1, priv->kbd_dev, priv->mouse_dev, gaminggear_macro);
	gaminggear_macro_free(gaminggear_macro);
}

static void shortcut_cb(PyraEventhandlerChannel *channel, guchar key_number, gpointer user_data) {
	PyraEventhandler *eventhandler = PYRA_EVENTHANDLER(user_data);
	PyraEventhandlerPrivate *priv = eventhandler->priv;
	PyraRMPMacroKeyInfo *macro_key_info;

	macro_key_info = pyra_rmp_get_macro_key_info(priv->rmp[priv->actual_profile_index], key_number - 1);
	pyra_play_shortcut(priv->kbd_dev, priv->mouse_dev, macro_key_info);
	g_free(macro_key_info);
}

static void quicklaunch_cb(PyraEventhandlerChannel *channel, guchar key_number, gpointer user_data) {
	PyraEventhandler *eventhandler = PYRA_EVENTHANDLER(user_data);
	PyraEventhandlerPrivate *priv = eventhandler->priv;
	PyraRMPMacroKeyInfo *macro_key_info;

	macro_key_info = pyra_rmp_get_macro_key_info(priv->rmp[priv->actual_profile_index], key_number - 1);
	pyra_play_quicklaunch_forked(macro_key_info);
	g_free(macro_key_info);
}

static gboolean profile_data_load(PyraEventhandler *eventhandler, GError **error) {
	PyraEventhandlerPrivate *priv = eventhandler->priv;
	guint i;

	for (i = 0; i < PYRA_PROFILE_NUM; ++i) {
		priv->rmp[i] = pyra_rmp_load(priv->device, i, error);
		if (priv->rmp[i] == NULL)
			return FALSE;
	}
	return TRUE;
}

static void device_add_cb(RoccatDeviceScannerInterface *interface, RoccatDevice *device, gpointer user_data) {
	PyraEventhandler *eventhandler = PYRA_EVENTHANDLER(user_data);
	PyraEventhandlerPrivate *priv = eventhandler->priv;
	GError *error = NULL;

	if (priv->device) {
		g_critical("Multiple devices present.");
		return;
	}

	priv->device = device;
	g_object_ref(G_OBJECT(device));
	roccat_device_debug(device);

	configuration_reload(eventhandler);

	/* Notificator gets initialized only once if a device is found.
	 * Gets deinited once in finalizer.
	 */
	if (!priv->profile_note)
		priv->profile_note = roccat_notification_profile_new(roccat_eventhandler_host_get_notificator(priv->host), PYRA_DEVICE_NAME);
	if (!priv->cpi_note)
		priv->cpi_note = roccat_notification_cpi_new(roccat_eventhandler_host_get_notificator(priv->host), PYRA_DEVICE_NAME);
	
	priv->actual_profile_index = pyra_actual_profile_read(priv->device, &error);
	if (error) {
		g_warning("Error reading actual profile: %s", error->message);
		g_clear_error(&error);
	}
	
	priv->kbd_dev = roccat_device_get_keyboard(priv->device, &error);
	if (priv->kbd_dev == -1) {
		g_critical("Can't play keyboard events: %s", error->message);
		g_clear_error(&error);
	}

	priv->mouse_dev = roccat_device_get_mouse(priv->device, &error);
	if (priv->mouse_dev == -1) {
		g_critical("Can't play mouse events: %s", error->message);
		g_clear_error(&error);
	}

	profile_data_load(eventhandler, &error);
	if (error) {
		g_warning("Error reading profile data: %s", error->message);
		g_clear_error(&error);
	}

	// FIXME create in init and just disconnect/reconnect?
	priv->dbus_server = pyra_dbus_server_new();
	pyra_dbus_server_connect(priv->dbus_server);
	g_signal_connect(G_OBJECT(priv->dbus_server), "open-gui", G_CALLBACK(open_gui_cb), eventhandler);
	g_signal_connect(G_OBJECT(priv->dbus_server), "profile-changed-outside", G_CALLBACK(profile_changed_cb), eventhandler);
	g_signal_connect(G_OBJECT(priv->dbus_server), "profile-data-changed-outside", G_CALLBACK(profile_data_changed_outside_cb), eventhandler);
	g_signal_connect(G_OBJECT(priv->dbus_server), "configuration-changed-outside", G_CALLBACK(configuration_changed_outside_cb), eventhandler);

	priv->active_window_changed_handler = g_signal_connect(G_OBJECT(priv->host), "active-window-changed", G_CALLBACK(window_changed_cb), eventhandler);

	priv->threads = gaminggear_macro_threads_new(PYRA_BUTTONS_NUM);
	
	pyra_eventhandler_channel_start(priv->channel, roccat_device_get_hidraw_endpoint_dev(priv->device, PYRA_INTERFACE_MOUSE), &error);
	if (error) {
		g_warning("Error starting chardev eventhandler: %s", error->message);
		g_clear_error(&error);
	}

	priv->device_set_up = TRUE;
}

static void stop(PyraEventhandler *eventhandler) {
	PyraEventhandlerPrivate *priv = eventhandler->priv;
	guint i;

	priv->device_set_up = FALSE;

	g_object_unref(priv->dbus_server);
	priv->dbus_server = NULL;

	pyra_eventhandler_channel_stop(priv->channel);

	g_signal_handler_disconnect(G_OBJECT(priv->host), priv->active_window_changed_handler);

	gaminggear_macro_threads_free(priv->threads);
	priv->threads = NULL;
	
	if (priv->device) {
		g_object_unref(G_OBJECT(priv->device));
		priv->device = NULL;
	}

	pyra_configuration_free(priv->config);
	priv->config = NULL;

	for (i = 0; i < PYRA_PROFILE_NUM; ++i) {
		pyra_rmp_free(priv->rmp[i]);
		priv->rmp[i] = NULL;
	}
}

static void device_remove_cb(RoccatDeviceScanner *scanner, RoccatDevice *device, gpointer user_data) {
	PyraEventhandler *eventhandler = PYRA_EVENTHANDLER(user_data);
	PyraEventhandlerPrivate *priv = eventhandler->priv;
	if (priv->device && gaminggear_device_equal(GAMINGGEAR_DEVICE(device), GAMINGGEAR_DEVICE(priv->device)))
		stop(eventhandler);
}

static void pyra_eventhandler_init(PyraEventhandler *eventhandler) {
	eventhandler->priv = PYRA_EVENTHANDLER_GET_PRIVATE(eventhandler);
}

static void finalize(GObject *object) {
	PyraEventhandler *eventhandler = PYRA_EVENTHANDLER(object);
	PyraEventhandlerPrivate *priv = eventhandler->priv;

	g_object_unref(priv->channel);
	priv->channel = NULL;

	if (priv->device_scanner) {
		g_object_unref(G_OBJECT(priv->device_scanner));
		priv->device_scanner = NULL;
	}

	stop(eventhandler);

	if (priv->profile_note) {
		roccat_notification_profile_free(priv->profile_note);
		priv->profile_note = NULL;
	}

	if (priv->cpi_note) {
		roccat_notification_cpi_free(priv->cpi_note);
		priv->cpi_note = NULL;
	}
	
	G_OBJECT_CLASS(pyra_eventhandler_parent_class)->finalize(object);
}

static GObject *constructor(GType gtype, guint n_properties, GObjectConstructParam *properties) {
	PyraEventhandler *eventhandler;
	PyraEventhandlerPrivate *priv;
	GObject *obj;

	obj = G_OBJECT_CLASS(pyra_eventhandler_parent_class)->constructor(gtype, n_properties, properties);
	eventhandler = PYRA_EVENTHANDLER(obj);
	priv = eventhandler->priv;

	priv->channel = pyra_eventhandler_channel_new();
	g_signal_connect(G_OBJECT(priv->channel), "cpi-changed", G_CALLBACK(cpi_changed_cb), eventhandler);
	g_signal_connect(G_OBJECT(priv->channel), "profile-changed", G_CALLBACK(profile_changed_cb), eventhandler);
	g_signal_connect(G_OBJECT(priv->channel), "quicklaunch", G_CALLBACK(quicklaunch_cb), eventhandler);
	g_signal_connect(G_OBJECT(priv->channel), "macro", G_CALLBACK(macro_cb), eventhandler);
	g_signal_connect(G_OBJECT(priv->channel), "shortcut", G_CALLBACK(shortcut_cb), eventhandler);

	priv->kbd_dev = -1;
	priv->mouse_dev = -1;

	priv->device_scanner = pyra_device_scanner_new();
	g_signal_connect(G_OBJECT(priv->device_scanner), "device-added", G_CALLBACK(device_add_cb), eventhandler);
	g_signal_connect(G_OBJECT(priv->device_scanner), "device-removed", G_CALLBACK(device_remove_cb), eventhandler);
	roccat_device_scanner_interface_start(ROCCAT_DEVICE_SCANNER_INTERFACE(priv->device_scanner));

	return obj;
}

static void set_property(GObject *object, guint prop_id, GValue const *value, GParamSpec *pspec) {
	PyraEventhandlerPrivate *priv = PYRA_EVENTHANDLER(object)->priv;

	switch(prop_id) {
	case PROP_HOST:
		priv->host = g_value_get_pointer(value);
		break;
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
	}
}

static void pyra_eventhandler_class_init(PyraEventhandlerClass *klass) {
	GObjectClass *gobject_class;

	gobject_class = G_OBJECT_CLASS(klass);

	gobject_class->constructor = constructor;
	gobject_class->finalize = finalize;
	gobject_class->set_property = set_property;

	g_type_class_add_private(klass, sizeof(PyraEventhandlerPrivate));

	g_object_class_install_property(gobject_class, PROP_HOST,
			g_param_spec_pointer("host",
					"Host",
					"Pointer to RoccatEventhandlerHost interface",
					G_PARAM_WRITABLE | G_PARAM_CONSTRUCT_ONLY));
}

ROCCAT_EVENTHANDLER_PLUGIN_API RoccatEventhandlerPlugin *roccat_eventhandler_plugin_new(RoccatEventhandlerHost *host) {
	return ROCCAT_EVENTHANDLER_PLUGIN(pyra_eventhandler_new(host));
}
